--   mousetrack.ms - a simple utility to exercise the mouseTrack() primitive
--
--   implements 3 tools: Spray for spraying geometry over a surface
--                       Select for selecting faces by painting on surface
--                       Sculpt for pushing and pulling a surface by painting
--
--     John Wainwright 1998
--

-- notes on the mouseTrack() function
--
--  Implemented in the "mousTrak.dlx" MAXScript plug-in found in the Bonus directory on the R2.5 CD.
--  This file needs to be placed in one of the MAX plugin directories, such as 3dsmax25\Plugins\ for
--  the function to be available in MAXScript.
--
--  The function has the form:
--
--      mouseTrack [on:<node>] [prompt:"msg"] [snap:#2D|#3D] [trackCallback:fn|#(fn,arg)]
--
--  where:
--     on:<node>  optionally specifies a scene object to track on.  If you don't supply
--                an object, the mouse tracks on the current active grid.  Note that the
--        object surface tracker uses the SDK function IntersectRay() which is 
--        not reliably implemented on all object types.   Editable meshes always 
--        work OK, so you can convertToMesh() if needed.
--
--     prompt:"msg"  displays the prompt message in the MAX staus line
--
--     snap:#2D|#3D  relevant when not tracking on an object surface (no on:<node> specified)
--                   if snap mode is on and #2D is supplied, snaps to snap points on grid, 
--        if #3D is specied snaps to any near snap point in space.
--
--     trackCallback:fn|#(fn,arg)   is where you specify the function to be called as the mouse
--                                  is dragged over the active grid or object surface.  You can 
--        specify it is a single function or a function and an argument value in a two element 
--        array.  The latter form is useful if you have a common callback function for many tasks
--        and want to send in a parameter to control its operation for a particular use.  The 
--        function you give must be of the following form:
--
--		    fn callback_fn msg ir obj faceNum shift ctrl alt = ...
--
--        in other words, a scripted function of 7 arguments.  While the tracking is going on,
--        it is called whenever a mouse action occurs, such as a button click or a drag, etc.
--        The 'msg' argument is a message code that indicates what kind of action occured, and
--        can be one of:
--   
--             #freeMove    - means the mouse is moved without a button being pressed
--             #mousePoint  - means the left mouse button has just been pressed
--             #mouseMove   - means the mouse is being dragged with the left button down
--             #mouseAbout  - means the right mouse button was clicked, normally meaning cancel
--
--        The 'ir' argument is the grid or surface intersection normal Ray at the current mouse
--        position.  The ray has a .pos property giving the point in space of the normal and
--        a .dir property giving the normal's direction vector.
--  
--        The 'obj' property is the object being dragged over or undefined if no on:<node> 
--        was supplied.
--
--        The 'faceNum' property is the number of the face the mosue is over if the object
--        being tracked is an editable mesh, undefined otherwise.
--                   
--        The 'shift', 'ctrl' and 'alt' arguments are true or false depending on the down or
--        up state of the <shitf>, <ctrl> and <alt> keys on the keyboard.
--
--        The function should return the special value #continue to continue tracking or any
--        other value to halt tracking.  That value then becomes the result of the original 
--        mouseTrack() call.
--                   
--

utility mesh_painter "Mesh Painter"
(
	-- each button opens a panel with controls for the selected tool
	checkbutton spray "Spray" tooltip:"Spray geometry over surface" offset:[-5,0] across:3
	checkbutton sel "Select" tooltip:"Select faces by painting"
	checkbutton sculpt "Sculpt" tooltip:"Sculpt mesh surface" offset:[5,0]
	
	-- the geometry Spray tool panel
	rollout spray_ro "Geometry Spray"
	(
		local surf, geom, lastp = [0,0,0]
		
		label psl "Surface:" align:#left
		pickbutton pick_surf "none" width:100
		label pgl "Geometry to spray:" align:#left
		pickbutton pick_geom "none" width:100
	
		group "Spray control"
		(
			spinner spacing "Min. spacing:"
			spinner jitter "Jitter"
			checkbutton do_spray "Spray" width:100 offset:[0,4]
		)
		
		-- the spray mouseTrack callback, makes copies of the spray geom at the normal ray
		fn spray_track msg ir obj faceNum shift ctrl alt =
		(
			if msg == #mouseMove and ir != undefined and
			   distance ir.pos lastp > spacing.value + random -jitter.value jitter.value
			then
			(
				copy geom pos:ir.pos dir:ir.dir
				lastp = ir.pos
			)
			if msg != #mouseAbort then #continue
		)

		on pick_surf picked obj do (surf = obj; pick_surf.text = obj.name)
		on pick_geom picked obj do (geom = obj; pick_geom.text = obj.name)
		
		on do_spray changed state do 
		(
			-- enter spray mode if we have geom & surf
			if surf != undefined and geom != undefined then
				undo on mouseTrack on:surf trackCallBack:spray_track
			do_spray.checked = false
		)
	)
	
	-- face select tool panel
	rollout sel_ro "Face Select"
	(
		local surf
		label psl "Surface:" align:#left
		pickbutton pick_surf "none" width:100
		checkbutton do_sel "Select" tooltip:"press <alt> key to subtract from selection" offset:[0,4] width:100
		
		-- face select callback, add or substract currect face to selection
		fn sel_track msg ir obj faceNum shift ctrl alt =
		(
			if msg == #mouseMove and faceNum != undefined then
			(
				if not alt then
					setFaceSelection obj #(faceNum) keep:true
				else
				(
					local s = getFaceSelection obj,
					      i = findItem s faceNum
					if i > 0 then setFaceSelection obj (deleteItem s i)
				)
				update obj
			)
			if msg == #mouseAbort then ok else #continue
		)

		on pick_surf picked obj do (surf = obj; pick_surf.text = obj.name)
		
		on do_sel changed state do 
		(
			if surf != undefined then 
			(
				-- convert to a mesh, if not already, and put in modifier panel
				if classOf surf.baseObject != editable_mesh then convertToMesh surf
				select surf
				max modify mode
				subObjectLevel = 2
				-- enter select paint mode
				undo on mouseTrack on:surf trackCallBack:sel_track
			)
			do_sel.checked = false
		)
	)
	
	-- the surface sculpting tool panel
	rollout sculpt_ro "Sculpt Surface"
	(
		local surf, last_face
		
		label psl "Surface:" align:#left
		pickbutton pick_surf "none" width:100
		
		group "Sculpting control"
		(
			spinner depth "Depth:" range:[0, 100, 2]
			checkbutton do_sculpt "Sculpt" tooltip:"drag to pull; <shift>-drag push surface" width:100 offset:[0,4]
		)
		
		-- face sculpting callback, shifts current face in or out along normal 
		fn sculpt_track msg ir obj faceNum shift ctrl alt =
		(
			if msg == #mouseMove and faceNum != undefined and faceNum != last_face then
			(
				local f = getFace obj faceNum, dir = if shift then -ir.dir else ir.dir,
				      d = dir * depth.value
				setVert obj f.x ((getVert obj f.x) + d)
				setVert obj f.y ((getVert obj f.y) + d)
				setVert obj f.z ((getVert obj f.z) + d)
				last_face = faceNum
				update obj
			)
			if msg == #mouseAbort then ok else #continue
		)

		on pick_surf picked obj do (surf = obj; pick_surf.text = obj.name)
		
		on do_sculpt changed state do 
		(
			if surf != undefined then 
			(
				if classOf surf.baseObject != editable_mesh then convertToMesh surf
				last_face = undefined
				undo on mouseTrack on:surf trackCallBack:sculpt_track
			)
			do_sculpt.checked = false
		)
	)
	
	on spray changed state do 
	(
		-- open the spray rollout, close the others
		sel.checked = sculpt.checked = false
		removeRollout sel_ro
		removeRollout sculpt_ro
		(if state then addRollout else removeRollout) spray_ro
	)
	
	on sel changed state do 
	(
		-- open the selection rollout, close the others
		spray.checked = sculpt.checked = false
		removeRollout spray_ro
		removeRollout sculpt_ro
		(if state then addRollout else removeRollout) sel_ro
	)
	
	on sculpt changed state do 
	(
		-- open the sculpt rollout, close the others
		spray.checked = sel.checked = false
		removeRollout spray_ro
		removeRollout sel_ro
		(if state then addRollout else removeRollout) sculpt_ro
	)
	
	on mesh_painter close do 
	(
		removeRollout spray_ro
		removeRollout sel_ro
		removeRollout sculpt_ro
	)
)